@@ -1,19 +1,34 @@
+#    Copyright 2020 Connor Ferster
+
+#    Licensed under the Apache License, Version 2.0 (the "License");
+#    you may not use this file except in compliance with the License.
+#    You may obtain a copy of the License at
+
+#        http://www.apache.org/licenses/LICENSE-2.0
+
+#    Unless required by applicable law or agreed to in writing, software
+#    distributed under the License is distributed on an "AS IS" BASIS,
+#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+#    See the License for the specific language governing permissions and
+#    limitations under the License.
+
 from collections import deque
 import copy
 from dataclasses import dataclass
-from functools import singledispatch, wraps
+from functools import singledispatch
 import importlib
 import inspect
 import itertools
 import math
 import os
 import pathlib
+import re
 from typing import Any, Union, Optional, Tuple, List
 import pyparsing as pp
 
-# TODO: Do something with inequality checks
+# TODO: Provide better formatting for inequality checks and bools
 
-# Five basic line types
+# Six basic line types
 @dataclass
 class CalcLine:
     line: deque
@@ -21,6 +36,13 @@ class CalcLine:
     latex: str
 
 
+@dataclass
+class SymbolicLine:
+    line: deque
+    comment: str
+    latex: str
+
+
 @dataclass
 class ConditionalLine:
     condition: deque
@@ -55,7 +77,7 @@ class BlankLine:  # Attributes not used on BlankLine but still req'd
     latex: str
 
 
-# Four types of cell
+# Five types of cell
 @dataclass
 class CalcCell:
     source: str
@@ -64,6 +86,7 @@ class CalcCell:
     lines: deque
     latex_code: str
 
+    # Test
     def __repr__(self):
         return str(
             "CalcCell(\n" + f"source=\n{self.source}\n" + f"lines=\n{self.lines}\n"
@@ -71,17 +94,32 @@ class CalcCell:
 
 
 @dataclass
-class OutputCell:
+class ShortCalcCell:
     source: str
     calculated_results: dict
-    lines: deque
     precision: int
-    cols: int
+    lines: deque
     latex_code: str
 
+    # Test
     def __repr__(self):
         return str(
-            "OutputCell(\n" + f"source=\n{self.source}\n" + f"lines=\n{self.lines}\n"
+            "ShortCalcCell(\n" + f"source=\n{self.source}\n" + f"lines=\n{self.lines}\n"
+        )
+
+
+@dataclass
+class SymbolicCell:
+    source: str
+    calculated_results: dict
+    precision: int
+    lines: deque
+    latex_code: str
+
+    # Test
+    def __repr__(self):
+        return str(
+            "CalcCell(\n" + f"source=\n{self.source}\n" + f"lines=\n{self.lines}\n"
         )
 
 
@@ -93,7 +131,7 @@ class ParameterCell:
     precision: int
     cols: int
     latex_code: str
-
+    # Test
     def __repr__(self):
         return str(
             "ParametersCell(\n"
@@ -108,15 +146,26 @@ class LongCalcCell:
     calculated_results: dict
     lines: deque
     precision: int
-    cols: int
     latex_code: str
-
+    # Test
     def __repr__(self):
         return str(
             "LongCalcCell(\n" + f"source=\n{self.source}\n" + f"lines=\n{self.lines}\n"
         )
 
 
+def is_number(s: str) -> bool:
+    """
+    A basic helper function because Python str methods do not
+    have this ability...
+    """
+    try:
+        float(s)
+        return True
+    except:
+        return False
+
+
 # The renderer class ("output" class)
 class LatexRenderer:
     def __init__(self, python_code_str: str, results: dict):
@@ -124,6 +173,7 @@ class LatexRenderer:
         self.results = results
         self.precision = 3
 
+    # Test
     def set_precision(self, n=3):
         """Sets the precision (number of decimal places) in all
         latex printed values. Default = 3"""
@@ -148,14 +198,15 @@ def latex(raw_python_source: str, calculated_results: dict, precision: int = 3)
 
 def categorize_raw_cell(
     raw_source: str, calculated_results: dict, precision: int = 3
-) -> Union[ParameterCell, OutputCell, CalcCell]:
+) -> Union[ParameterCell, CalcCell]:
     """
     Return a "Cell" type depending on the source of the cell.
     """
     raw_source = raw_source.strip()
     if test_for_parameter_cell(raw_source):
+        comment_tag_removed = strip_cell_code(raw_source)
         cell = ParameterCell(
-            source=strip_cell_code(raw_source),
+            source=comment_tag_removed,
             calculated_results=calculated_results,
             lines=deque([]),
             precision=precision,
@@ -163,23 +214,34 @@ def categorize_raw_cell(
             latex_code="",
         )
 
-    elif test_for_output_cell(raw_source):
-        cell = OutputCell(
-            source=strip_cell_code(raw_source),
+    elif test_for_long_cell(raw_source):
+        comment_tag_removed = strip_cell_code(raw_source)
+        cell = LongCalcCell(
+            source=comment_tag_removed,
             calculated_results=calculated_results,
             lines=deque([]),
             precision=precision,
-            cols=1,
             latex_code="",
         )
 
-    elif test_for_long_cell(raw_source):
-        cell = LongCalcCell(
-            source=strip_cell_code(raw_source),
+    elif test_for_short_cell(raw_source):
+        comment_tag_removed = strip_cell_code(raw_source)
+        cell = ShortCalcCell(
+            source=comment_tag_removed,
+            calculated_results=calculated_results,
+            lines=deque([]),
+            precision=precision,
+            latex_code="",
+        )
+
+    # Test: a symbolic calc
+    elif test_for_symbolic_cell(raw_source):
+        comment_tag_removed = strip_cell_code(raw_source)
+        cell = SymbolicCell(
+            source=comment_tag_removed,
             calculated_results=calculated_results,
             lines=deque([]),
             precision=precision,
-            cols=1,
             latex_code="",
         )
 
@@ -209,8 +271,8 @@ def strip_cell_code(raw_source: str) -> str:
 
 
 def categorize_lines(
-    cell: Union[CalcCell, ParameterCell, OutputCell]
-) -> Union[CalcCell, ParameterCell, OutputCell]:
+    cell: Union[CalcCell, ParameterCell]
+) -> Union[CalcCell, ParameterCell]:
     """
     Return 'cell' with the line data contained in cell_object.source categorized
     into one of four types:
@@ -221,16 +283,17 @@ def categorize_lines(
     categorize_lines(calc_cell) is considered the default behaviour for the 
     singledispatch categorize_lines function.
     """
-    # source = cell.source.rstrip()
     outgoing = cell.source.rstrip().split("\n")
     incoming = deque([])
     calculated_results = cell.calculated_results
     override = ""
     for line in outgoing:
-        if isinstance(cell, (ParameterCell, OutputCell)):
+        if isinstance(cell, ParameterCell):
             override = "parameter"
         elif isinstance(cell, LongCalcCell):
             override = "long"
+        elif isinstance(cell, SymbolicCell):
+            override = "symbolic"
         categorized = categorize_line(line, calculated_results, override)
         categorized_w_result_appended = add_result_values_to_line(
             categorized, calculated_results
@@ -248,7 +311,7 @@ def categorize_line(
     fits the appropriate criteria. Raise ValueError, otherwise.
 
     'override' is a str used to short-cut the tests in categorize_line(). e.g.
-    if the cell that the lines belong to is a ParameterCell or OutputCell, 
+    if the cell that the lines belong to is a ParameterCell,
     we do not need to run the test_for_parameter_line() function on the line
     because, in a ParameterCell, all lines will default to a ParameterLine
     because of the cell it's in and how that cell is supposed to behave.
@@ -264,7 +327,7 @@ def categorize_line(
     categorized_line = None
 
     # Override behaviour
-    if not test_for_blank_line(line):
+    if not test_for_blank_line(line):  # True is a blank line
         if override == "parameter":
             categorized_line = ParameterLine(
                 split_parameter_line(line, calculated_results), comment, ""
@@ -272,13 +335,29 @@ def categorize_line(
             return categorized_line
 
         elif override == "long":
-            categorized_line = LongCalcLine(code_reader(line), comment, "")
+            if test_for_parameter_line(
+                line
+            ):  # A parameter can exist in a long cell, too
+                categorized_line = ParameterLine(
+                    split_parameter_line(line, calculated_results), comment, ""
+                )
+            else:
+                categorized_line = LongCalcLine(code_reader(line), comment, "")
+            return categorized_line
+        elif override == "symbolic":
+            categorized_line = SymbolicLine(code_reader(line), comment, "")
+            return categorized_line
+        elif override == "short":
+            categorized_line = CalcLine(code_reader(line), comment, "")
             return categorized_line
         elif True:
-            pass  # Future override conditions can be put here
+            pass  # Future override conditions to match new cell types can be put here
 
     # Standard behaviour
-    if test_for_parameter_line(line):
+    if line == "\n" or line == "":
+        categorized_line = BlankLine(line, "", "")
+
+    elif test_for_parameter_line(line):
         categorized_line = ParameterLine(
             split_parameter_line(line, calculated_results), comment, ""
         )
@@ -306,20 +385,19 @@ def categorize_line(
     elif "=" in line:
         categorized_line = CalcLine(code_reader(line), comment, "")
 
-    elif len(line) == 1:
+    elif len(expr_parser(line)) == 1:
         categorized_line = ParameterLine(
             split_parameter_line(line, calculated_results), comment, ""
         )
 
     else:
-        if line == "\n" or line == "":
-            categorized_line = BlankLine(line, "", "")
-        else:
-            raise ValueError(
-                f"Line: {line} is not recognized for rendering.\n"
-                "Lines must include either 'if/else' or '='."
-                "Did you intend to create an '# Output' cell?"
-            )
+        raise ValueError(
+            f"Line: {line} is not recognized for rendering.\n"
+            "Lines must either:\n"
+            "\t * Be the name of a previously assigned single variable\n"
+            "\t * Be an arithmetic variable assignment (i.e. calculation that uses '=' in the line)\n"
+            "\t * Be a conditional arithmetic assignment (i.e. uses 'if', 'elif', or 'else', each on a single line)"
+        )
     return categorized_line
 
 
@@ -359,6 +437,11 @@ def results_for_conditionline(line_object, calculated_results: dict):
     return line_object
 
 
+@add_result_values_to_line.register(SymbolicLine)
+def results_for_symbolicline(line_object, calculated_results):
+    return line_object
+
+
 @add_result_values_to_line.register(BlankLine)
 def results_for_blank(line_object, calculated_results):
     return line_object
@@ -366,6 +449,13 @@ def results_for_blank(line_object, calculated_results):
 
 @singledispatch
 def convert_cell(cell_object):
+    """
+    Return the cell_object with all of its lines run through the function, 
+    'convert_lines()', effectively converting each python element in the parsed
+    deque in the equivalent element in latex. 
+
+    The result remains stored in cell.lines
+    """
     raise TypeError(
         f"Cell object {type(cell_object)} is not yet recognized in convert_cell()"
     )
@@ -382,8 +472,19 @@ def convert_calc_cell(cell: CalcCell) -> CalcCell:
     return cell
 
 
+@convert_cell.register(ShortCalcCell)
+def convert_calc_cell(cell: ShortCalcCell) -> ShortCalcCell:
+    outgoing = cell.lines
+    calculated_results = cell.calculated_results
+    incoming = deque([])
+    for line in outgoing:
+        incoming.append(convert_line(line, calculated_results))
+    cell.lines = incoming
+    return cell
+
+
 @convert_cell.register(LongCalcCell)
-def convert_calc_cell(cell: LongCalcCell) -> LongCalcCell:
+def convert_longcalc_cell(cell: LongCalcCell) -> LongCalcCell:
     outgoing = cell.lines
     calculated_results = cell.calculated_results
     incoming = deque([])
@@ -404,8 +505,8 @@ def convert_parameter_cell(cell: ParameterCell) -> ParameterCell:
     return cell
 
 
-@convert_cell.register(OutputCell)
-def convert_parameter_cell(cell: OutputCell) -> OutputCell:
+@convert_cell.register(SymbolicCell)
+def convert_symbolic_cell(cell: SymbolicCell) -> SymbolicCell:
     outgoing = cell.lines
     calculated_results = cell.calculated_results
     incoming = deque([])
@@ -417,9 +518,19 @@ def convert_parameter_cell(cell: OutputCell) -> OutputCell:
 
 @singledispatch
 def convert_line(
-    line_object: Union[CalcLine, ConditionalLine, ParameterLine],
+    line_object: Union[
+        CalcLine, ConditionalLine, ParameterLine, SymbolicLine, BlankLine
+    ],
     calculated_results: dict,
-) -> Union[CalcLine, ConditionalLine, ParameterLine]:
+) -> Union[CalcLine, ConditionalLine, ParameterLine, SymbolicLine, BlankLine]:
+    """
+    Returns 'line_object' with its .line attribute converted into a 
+    deque with elements that have been converted to their appropriate
+    Latex counterparts.
+
+    convert_line() runs the deque through all of the conversion functions
+    as organized in `swap_calculation()`.
+    """
     raise TypeError(
         f"Cell object {type(line_object)} is not yet recognized in convert_line()"
     )
@@ -471,13 +582,21 @@ def convert_parameter(line, calculated_results):
     return line
 
 
+@convert_line.register(SymbolicLine)
+def convert_symbolic_line(line, calculated_results):
+    line.line = swap_symbolic_calcs(line.line)
+    return line
+
+
 @convert_line.register(BlankLine)
 def convert_blank(line, calculated_results):
     return line
 
 
 @singledispatch
-def format_cell(cell_object: Union[OutputCell, ParameterCell, CalcCell]):
+def format_cell(
+    cell_object: Union[ParameterCell, LongCalcCell, CalcCell, SymbolicCell]
+) -> Union[ParameterCell, LongCalcCell, CalcCell, SymbolicCell]:
     raise TypeError(
         f"Cell type {type(cell_object)} has not yet been implemented in format_cell()."
     )
@@ -511,8 +630,8 @@ def parameters_cell(cell: ParameterCell):
             line.latex = latex_param.replace("=", "&=")
 
     latex_block = " ".join(
-        [line.latex for line in cell.lines]
-    ).rstrip()  # .rstrip(): Hack to solve another problem
+        [line.latex for line in cell.lines if not isinstance(line, BlankLine)]
+    ).rstrip()  # .rstrip(): Hack to solve another problem of empty lines in {aligned} environment
     cell.latex_code = "\n".join([opener, begin, latex_block, end, closer])
     return cell
 
@@ -540,6 +659,28 @@ def format_calc_cell(cell: CalcCell) -> str:
     return cell
 
 
+@format_cell.register(ShortCalcCell)
+def format_shortcalc_cell(cell: ShortCalcCell) -> str:
+    line_break = "\\\\[10pt]\n"
+    precision = cell.precision
+    incoming = deque([])
+    for line in cell.lines:
+        line = round_and_render_line_objects_to_latex(line, precision)
+        line = format_lines(line)
+        incoming.append(line)
+    cell.lines = incoming
+
+    latex_block = line_break.join([line.latex for line in cell.lines if line.latex])
+    opener = "\\["
+    begin = "\\begin{aligned}"
+    end = "\\end{aligned}"
+    closer = "\\]"
+    cell.latex_code = "\n".join([opener, begin, latex_block, end, closer]).replace(
+        "\n" + end, end
+    )
+    return cell
+
+
 @format_cell.register(LongCalcCell)
 def format_longcalc_cell(cell: LongCalcCell) -> str:
     line_break = "\\\\[10pt]\n"
@@ -563,14 +704,18 @@ def format_longcalc_cell(cell: LongCalcCell) -> str:
     return cell
 
 
-@format_cell.register(OutputCell)
-def format_output_cell(cell: OutputCell) -> str:
+@format_cell.register(SymbolicCell)
+def format_symbolic_cell(cell: SymbolicCell) -> str:
     line_break = "\\\\[10pt]\n"
     precision = cell.precision
+    incoming = deque([])
     for line in cell.lines:
         line = round_and_render_line_objects_to_latex(line, precision)
         line = format_lines(line)
-    latex_block = line_break.join([line.latex for line in cell.lines])
+        incoming.append(line)
+    cell.lines = incoming
+
+    latex_block = line_break.join([line.latex for line in cell.lines if line.latex])
     opener = "\\["
     begin = "\\begin{aligned}"
     end = "\\end{aligned}"
@@ -584,7 +729,16 @@ def format_output_cell(cell: OutputCell) -> str:
 @singledispatch
 def round_and_render_line_objects_to_latex(
     line: Union[CalcLine, ConditionalLine, ParameterLine], precision: int
-):
+):  # Not called for symbolic lines; see format_symbolic_cell()
+    """
+    Returns 'line' with the elements of the deque in its .line attribute
+    converted into their final string form for rendering (thereby preserving
+    its intermediate step) and populates the 
+    .latex attribute with the joined string from .line.
+
+    'precision' is the number of decimal places that each object should
+    be rounded to for display.
+    """
     raise TypeError(
         f"Line type {type(line)} not recognized yet in round_and_render_line_objects_to_latex()"
     )
@@ -630,13 +784,23 @@ def round_and_render_conditional(
     return line
 
 
+@round_and_render_line_objects_to_latex.register(SymbolicLine)
+def round_and_render_symbolic(line: SymbolicLine, precision: int) -> SymbolicLine:
+    rounded_line = round_and_render(line.line, precision)
+    line.line = rounded_line
+    line.latex = " ".join(rounded_line)
+    return line
+
+
 @round_and_render_line_objects_to_latex.register(BlankLine)
 def round_and_render_blank(line, precision):
     return line
 
 
 @singledispatch
-def convert_applicable_long_lines(line: Union[ConditionalLine, CalcLine]):
+def convert_applicable_long_lines(
+    line: Union[ConditionalLine, CalcLine]
+):  # Not called for symbolic lines; see format_symbolic_cell()
     raise TypeError(
         f"Line type {type(line)} not yet implemented in convert_applicable_long_lines()."
     )
@@ -786,6 +950,14 @@ def convert_calc_line_to_long(calc_line: CalcLine) -> LongCalcLine:
 
 @singledispatch
 def format_lines(line_object):
+    """
+    format_lines adds small, context-dependent pieces of latex code in
+    amongst the latex string in the line_object.latex attribute. This involves
+    things like inserting "&" or linebreak characters for equation alignment, 
+    formatting comments stored in the .comment attribute and putting them at 
+    the end of the calculation, or the distinctive "Since, <condition> ..."
+    text that occurs when a conditional calculation is rendered.
+    """
     raise TypeError(
         f"Line type {type(line_object)} is not yet implemented in format_lines()."
     )
@@ -815,34 +987,22 @@ def format_conditional_line(line: ConditionalLine) -> ConditionalLine:
         latex_condition = " ".join(line.true_condition)
         a = "{"
         b = "}"
-        new_math_env = "\n\\end{aligned}\n\\]\n\\[\n\\begin{aligned}\n"
-        first_line = f"&\\text{a}Since, {b}{latex_condition}:{new_math_env}"
-        if line.condition_type == "else":
-            first_line = ""
-        transition_line = ""  # "&\\text{Therefore:} \\\\"
-        line_break = "\\\\\n"
         comment_space = ""
         comment = ""
         if line.comment:
             comment_space = "\\;"
             comment = format_strings(line.comment, comment=True)
+        new_math_env = "\n\\end{aligned}\n\\]\n\\[\n\\begin{aligned}\n"
+        first_line = f"&\\text{a}Since, {b}{latex_condition}:{comment_space}{comment}{new_math_env}"
+        if line.condition_type == "else":
+            first_line = ""
+        line_break = "\\\\\n"
+
         outgoing = deque([])
         for calc_line in line.true_expressions:
             outgoing.append((format_lines(calc_line)).latex)
         latex_exprs = line_break.join(outgoing)
-        if len(outgoing) > 1:
-            line.latex = (
-                first_line
-                + comment_space
-                + comment
-                + transition_line
-                + line_break
-                + latex_exprs
-            )
-        else:
-            line.latex = (
-                first_line + comment_space + comment + transition_line + latex_exprs
-            )
+        line.latex = first_line + latex_exprs
         return line
     else:
         line.latex = ""
@@ -856,8 +1016,8 @@ def format_long_calc_line(line: LongCalcLine) -> LongCalcLine:
     for positioning within the "\aligned" latex environment.
     """
     latex_code = line.latex
-    long_latex = latex_code.replace("=", "\\\\&=")  # Change all
-    long_latex = long_latex.replace("\\\\&=", "&=", 1)  # Fix the first one
+    long_latex = latex_code.replace("=", "\\\\&=")  # Change all...
+    long_latex = long_latex.replace("\\\\&=", "&=", 1)  # ...except the first one
     line_break = "\\\\\n"
     comment_space = ""
     comment = ""
@@ -877,6 +1037,15 @@ def format_param_line(line: ParameterLine) -> ParameterLine:
     return line
 
 
+@format_lines.register(SymbolicLine)
+def format_symbolic_line(line: SymbolicLine) -> SymbolicLine:
+    replaced = line.latex.replace("=", "&=")
+    comment_space = "\\;"
+    comment = format_strings(line.comment, comment=True)
+    line.latex = f"{replaced}{comment_space}{comment}\n"
+    return line
+
+
 @format_lines.register(BlankLine)
 def format_blank_line(line: BlankLine) -> BlankLine:
     line.latex = ""
@@ -901,9 +1070,6 @@ def split_conditional(line: str, calculated_results):
     expr_acc = deque([])
     for line in expr_deque:
         categorized = categorize_line(line, calculated_results)
-        # categorized_w_result_appended = add_result_values_to_line(
-        #     categorized, calculated_results
-        # )
         expr_acc.append(categorized)
 
     return (
@@ -933,7 +1099,7 @@ def test_for_parameter_line(line: str) -> bool:
             expr_as_code = code_reader(line)
             if len(expr_as_code) == 3 and expr_as_code[1] == "=":  # Param = value
                 return True
-            elif (
+            elif (  # Special check required for -ve nums
                 len(expr_as_code) == 4
                 and expr_as_code[1] == "="
                 and expr_as_code[2] == "-"
@@ -956,17 +1122,6 @@ def test_for_parameter_cell(raw_python_source: str) -> bool:
     return False
 
 
-def test_for_output_cell(raw_python_source: str) -> bool:
-    """
-    Returs True if the text "# Out" is in the first line of 
-    'raw_python_source'. False otherwise.
-    """
-    first_element = raw_python_source.split("\n")[0]
-    if "#" in first_element and "out" in first_element.lower():
-        return True
-    return False
-
-
 def test_for_long_cell(raw_python_source: str) -> bool:
     """
     Returns True if the text "# Long" is in the first line of
@@ -978,6 +1133,28 @@ def test_for_long_cell(raw_python_source: str) -> bool:
     return False
 
 
+def test_for_short_cell(raw_python_source: str) -> bool:
+    """
+    Returns True if the text "# Long" is in the first line of
+    `raw_python_source`. False otherwise.
+    """
+    first_element = raw_python_source.split("\n")[0]
+    if "#" in first_element and "short" in first_element.lower():
+        return True
+    return False
+
+
+def test_for_symbolic_cell(raw_python_source: str) -> bool:
+    """
+    Returns True if the text "# Long" is in the first line of
+    `raw_python_source`. False otherwise.
+    """
+    first_element = raw_python_source.split("\n")[0]
+    if "#" in first_element and "symbolic" in first_element.lower():
+        return True
+    return False
+
+
 def test_for_blank_line(source: str) -> bool:
     """
     Returns True if 'source' is effectively a blank line, 
@@ -987,6 +1164,13 @@ def test_for_blank_line(source: str) -> bool:
     return not bool(source.strip())
 
 
+def test_for_conditional_line(source: str) -> bool:
+    """
+    Returns True if 'source' appears to be conditional expression.
+    """
+    return ":" in source and ("if" in source or "else" in source)
+
+
 def test_for_single_dict(source: str, calc_results: dict) -> bool:
     """
     Returns True if 'source' is a str representing a variable name
@@ -1039,6 +1223,11 @@ def round_and_render(line_of_code: deque, precision: int) -> deque:
     """
     outgoing = deque([])
     for item in line_of_code:
+        if hasattr(item, "__len__") and not isinstance(item, (str, dict)):
+            try:
+                rounded = [round(v, precision) for v in item]
+            except:
+                rounded = item
         if not isinstance(item, (str, int)):
             try:
                 rounded = round(item, precision)  # Rounding
@@ -1055,26 +1244,28 @@ def latex_repr(item: Any) -> str:
     Return a str if the object, 'item', has a special repr method
     for rendering itself in latex. If not, returns str(result).
     """
-    if hasattr(item, "hc_latex"):
-        try:
-            return item.hc_latex().replace("$", "")
-        except TypeError:
-            return str(item)
-
-    elif hasattr(item, "_repr_latex_"):
+    if hasattr(item, "_repr_latex_"):
         return item._repr_latex_().replace("$", "")
 
     elif hasattr(item, "latex"):
         try:
             return item.latex().replace("$", "")
         except TypeError:
-            str(item)
+            return str(item)
 
     elif hasattr(item, "to_latex"):
         try:
             return item.to_latex().replace("$", "")
         except TypeError:
-            str(item)
+            return str(item)
+
+    elif hasattr(item, "__len__") and not isinstance(item, (str, dict, tuple)):
+        comma_space = ",\\ "
+        try:
+            array = "[" + comma_space.join([str(v) for v in item]) + "]"
+            return array
+        except TypeError:
+            return str(item)
 
     else:
         return str(item)
@@ -1314,6 +1505,9 @@ def eval_conditional(conditional_str: str, **kwargs) -> str:
     # unpacking-a-passed-dictionary-into-the-functions-name-space-in-python
     exec(",".join(kwargs) + ", = kwargs.values()")
     try:
+        # It would be good to sanitize the code coming in on 'conditional_str'
+        # Should this code be forced into using only boolean operators?
+        # Do not need to cross this bridge, yet.
         return eval(conditional_str)
     except SyntaxError:
         return conditional_str
@@ -1338,13 +1532,11 @@ def expr_parser(expr_as_str: str) -> deque:
     """
     term = pp.Word(pp.srange("[A-Za-z0-9_'\".]"), pp.srange("[A-Za-z0-9_'\".]"))
     operator = pp.Word("+-*/^%<>=~!,")
-    # eol = pp.Word(";")
     func = pp.Word(pp.srange("[A-Za-z0-9_.]")) + pp.FollowedBy(
         pp.Word(pp.srange("[A-Za-z0-9_()]"))
     )
     string = pp.Word(pp.srange("[A-Za-z0-9'\"]"))
     group = term ^ operator ^ func ^ string
-    # parenth = pp.nestedExpr(content=group ^ eol)
     parenth = pp.nestedExpr(content=group)
     master = group ^ parenth
     expr = pp.OneOrMore(master)
@@ -1468,20 +1660,16 @@ def swap_math_funcs(pycode_as_deque: deque) -> deque:
     for index, item in enumerate(pycode_as_deque):
         next_idx = min(index + 1, length - 1)
         next_item = pycode_as_deque[next_idx]
-        if type(item) is deque:
+        if isinstance(item, deque):
             new_item = swap_math_funcs(item)  # recursion!
             swapped_deque.append(new_item)
             if close_bracket_token:
                 swapped_deque.append(b)
                 close_bracket_token = False
-        elif close_bracket_token:
-            swapped_deque.append(latex_math_funcs.get(item, item))
-            new_item = f"{b}"
-            close_bracket_token = False
-            swapped_deque.append(new_item)
         elif (
             isinstance(next_item, deque)
-            and item.isalpha()
+            and isinstance(item, str)
+            and re.match(r"^[A-Za-z_]+$", item)
             and item not in latex_math_funcs
         ):
             ops = "\\operatorname"
@@ -1536,9 +1724,7 @@ def swap_superscripts(pycode_as_deque: deque) -> deque:
         if isinstance(item, deque) and not close_bracket_token:
             new_item = swap_superscripts(item)  # recursion!
             pycode_with_supers.append(new_item)
-        elif not isinstance(next_item, deque) and "**" in str(
-            next_item
-        ):  # next_item == "**":#"**" in str(next_item):
+        elif not isinstance(next_item, deque) and "**" in str(next_item):
             pycode_with_supers.append(l_par)
             pycode_with_supers.append(item)
             pycode_with_supers.append(r_par)
@@ -1550,7 +1736,7 @@ def swap_superscripts(pycode_as_deque: deque) -> deque:
             close_bracket_token
             and not isinstance(prev_item, deque)
             and "**" in str(prev_item)
-        ):  # prev_item == "**":
+        ):
             pycode_with_supers.append(item)
             new_item = f"{b}"
             pycode_with_supers.append(new_item)
@@ -1566,7 +1752,8 @@ def swap_for_greek(pycode_as_deque: deque) -> deque:
     Returns full line of code as deque with any Greek terms swapped in for words describing
     Greek terms, e.g. 'beta' -> 'β'
     """
-    # "eta" and "psi" need to be last on the list b/c they are substrings of "theta" and "epsilon"
+    # "eta" and "psi" need to be last on the list b/c they are substrings
+    # of "theta" and "epsilon"
     greeks = [
         "alpha",
         "beta",
@@ -1598,6 +1785,8 @@ def swap_for_greek(pycode_as_deque: deque) -> deque:
             new_item = swap_for_greek(item)  # recursion!
             pycode_with_greek.append(new_item)
         elif isinstance(item, str):
+            # This code is intended to swap out the greeks but to also swap
+            # out greeks in subscripts or sub-sub-scripts, etc.
             for greek in greeks:
                 if (greek in item or greek.capitalize() in item) and (
                     greek not in greek_exceptions
@@ -1626,8 +1815,6 @@ def swap_for_greek(pycode_as_deque: deque) -> deque:
                     )
 
             pycode_with_greek.append(item)
-            # else:
-            #     pycode_with_greek.append(item)
         else:
             pycode_with_greek.append(item)
     return pycode_with_greek
@@ -1644,52 +1831,8 @@ def swap_values(pycode_as_deque: deque, tex_results: dict) -> deque:
         if isinstance(item, deque):
             outgoing.append(swap_values(item, tex_results))
         else:
-            # try:
             swapped_value = tex_results.get(item, item)
             if isinstance(swapped_value, str) and swapped_value != item:
                 swapped_value = format_strings(swapped_value, comment=False)
             outgoing.append(swapped_value)
-            # except:
-            #     if isinstance(swapped_value, str) and swapped_value != item:
-            #         swapped_value = format_strings(swapped_value, comment=False)
-            #         swapped_values.append(swapped_value)
     return outgoing
-
-## Basic Usage 2: Decorator
-
-def handcalc(func):
-    #@wraps(func)
-    def wrapper(*args, **kwargs):
-        func_source = inspect.getsource(func)
-        cell_source = func_source_to_cell(func_source)
-        calculated_results = func(*args, **kwargs) # Func must use `return locals()`
-        if not isinstance(calculated_results, dict):
-            raise ValueError(f"Return value of decorated function should be locals(),",
-                             " not {calculated_results}")
-        renderer = LatexRenderer(cell_source, calculated_results)
-        latex_code = renderer.render()
-
-        latex_code = latex_code.replace("\\[","", 1).replace("\\]","")
-        return (latex_code, calculated_results)
-    return wrapper
-        
-def func_source_to_cell(source: str):
-    """
-    Returns a string that represents `source` but with no signature, doc string,
-    or return statement.
-    
-    `source` is a string representing a function's complete source code.
-    """
-    source_lines = source.split('\n')
-    acc = []
-    for line in source_lines:
-        doc_string = False
-        if not doc_string and '"""' in line: 
-            doc_string = True
-            continue
-        elif doc_string and '"""' in line: 
-            doc_string = False
-            continue
-        if "def" not in line and not doc_string and "return" not in line and "@" not in line:
-            acc.append(line)
-    return "\n".join(acc)
